#if defined __APPLE__
#include "SamplesOSX.h"
#else
#include "sample.h"
#endif
#include "sample_common.h"
#include "sprk.h"

#include "sprk_exchange.h"

#include <iostream>
#include <functional>
#include <fstream>

using namespace HPS;

void Sample::Exchange_Component_Interaction(WindowKey wk, Canvas canvas, View view, Model model, SegmentKey modelKey)
{
	try
	{
	/* Sample: Exchange_Component_Interaction
	 * This sample will show you how to:
	 * - Import and Exchange file and check for exceptions
	 * - Traverse the component structure of a CADModel
	 * - Obtain a component from a selection event
	 * - Highlight a selected component
	 * - Define named styles and glyphs
	 * 
	 * Read more about the topics covered in this sample in our Programming Guide 
	 * - Integration with HOOPS Exchange
	 * - Section 6.1:		Selection
	 * - Section 6.2:		Highlighting
	 * - Section 4.3:		Styles
	 * - Section 4.5:		Glyphs */
	
	std::ofstream file;

	/* NOTE:
	 * For this sample we are using a lambda to traverse the components of the
	 * Exchange file.
	 * The same result could be obtain by putting the body of the lambda in a 
	 * separate helper function */
	std::function<void(Component, int)> traverseComponents = 
		[&] (Component component, int depth)
	{
		UTF8 outputString;

		for(int i=0; i < depth; ++i)
			outputString = outputString + "  ";
		
		describeComponent(component, &outputString);

		StringMetadata metadata = component.GetMetadata("Name");
		if(metadata.Type() != Type::None)
			outputString = outputString + " - Name: " + metadata.GetValue();
		else
			outputString = outputString + " - Name: N/A";

		file << outputString << std::endl;

		ComponentArray cArray = component.GetSubcomponents();

		if(cArray.size() == 0)
			return;
		else
		{
			++depth;
			for(size_t i=0, e=cArray.size(); i < e; ++i)
			{
				traverseComponents(cArray[i], depth);
			}
		}
	};

	/* This next block of code shows you how to load an Exchange file
	 * 1. Create an ImportOptionsKit
	 * 2. Use the Exchange::File::Import method to load the desired file, with the options you specified
	 * 3. Since this operation happens on a separate thread, wait for it to complete before proceeding 
	 * 4. Use a try-catch block to capture any Exchange IO exceptions 
	 * 5. Obtain a CADModel object through the Exchange::ImportNotifier */
	Exchange::ImportNotifier exchangeNotifier;
	CADModel modelFile;
	try {
		UTF8 file_in = INPUT_FOLDER + "/revolved.prt";
		Exchange::ImportOptionsKit iok;
		iok.SetBRepData(true);
		exchangeNotifier = Exchange::File::Import(static_cast<const char*>(file_in), iok);
		exchangeNotifier.Wait();

		modelFile = exchangeNotifier.GetCADModel();
		view.Delete();
		view = modelFile.ActivateDefaultCapture();
		canvas.AttachViewAsLayout(view);

		modelKey = view.GetAttachedModel().GetSegmentKey();
		modelKey.InsertDistantLight(Vector(1, 1, 0));
		view.FitWorld();
	}

	catch(IOException ioe) {
		UTF8 error_message;
		error_message = "Exchange Import Unsuccessful";
		modelKey.InsertText(Point(0, 0, 0), static_cast<const char*>(error_message));
		view.FitWorld();
		view.Update();
	}

	/* This block of code uses the lambda defined above to recursively walk the component
	 * structure of the imported file, and output it to a text file.
	 * 
	 * For each component, its type and its name, when available, will be printed to the
	 * output file.
	 * 
	 * The type of a component can be retrieved using the GetComponentType.
	 * The name of a component is one of the metadata that can be associated with a 
	 * component. To obtain it, simply call the GetMetadata method with "Name" as 
	 * the argument.
	 * 
	 * If the type of the returned metadata is 'None' it means that the component
	 * did not have a name metadata set on it */
	ComponentArray root = modelFile.GetSubcomponents();
	UTF8 file_out = OUTPUT_FOLDER + "/exchange_component_interaction.txt";
	
	file.open(static_cast<const char *>(file_out), std::ios::out);
	traverseComponents(root[0], 0);
	file.close();

	/* This block sets up the scene used in this sample.
	 * A segment is created which will be used to output text to the screen, and
	 * another segment is created to be used to house a marker which will indicate the
	 * location of the selection operation */
	SegmentKey elements = modelKey.Subsegment();
	elements.GetCameraControl().SetPosition(Point(0, 0, 5)).
		SetTarget(Point(0, 0, 0)).
		SetUpVector(Vector(0, 1, 0)).
		SetField(2, 2).
		SetProjection(Camera::Projection::Stretched);

	SegmentKey description = elements.Subsegment();
	description.GetDrawingAttributeControl().SetDepthRange(0.0f, 0.1f);
	description.GetVisibilityControl().SetEverything(false).SetText(true);
	
	description.GetTextAttributeControl().SetSize(0.02f, Text::SizeUnits::SubscreenRelative).
		SetAlignment(Text::Alignment::Center, Text::ReferenceFrame::WorldAligned, Text::Justification::CenterJustified);
	description.InsertText(Point(0, -0.80f, 0), "A file describing the component structure of this model");
	description.InsertText(Point(0, -0.85f, 0), "has been saved here: ");
	description.InsertText(Point(0, -0.90f, 0), OUTPUT_FOLDER + "/exchange_component_interaction.txt");

	/* This block of code defines the glyph symbol that will be used to draw the marker
	 * and the named style which will be used to highlight the selected component */
	PortfolioKey mpKey, vpKey;
	modelKey.GetPortfolioControl().ShowTop(mpKey);
	mpKey.DefineGlyph("myGlyph", GlyphKit::GetDefault(Glyph::Default::Plus));

	view.GetSegmentKey().GetPortfolioControl().ShowTop(vpKey);
	SegmentKey highlightStyle = view.GetSegmentKey().Subsegment();
	highlightStyle.GetMaterialMappingControl().
		SetFaceColor(RGBAColor(1, 0.5f, 0, 1)).
		SetEdgeColor(RGBAColor(1, 0.5f, 0, 1));
	vpKey.DefineNamedStyle("orangeHighlight", highlightStyle);

	MarkerKit markerKit;
	markerKit.SetPoint(Point(-0.05f, -0.1f, 0));
	SegmentKey marker = elements.Subsegment();
	marker.GetMarkerAttributeControl().SetSymbol("myGlyph").SetSize(3);
	marker.InsertMarker(markerKit);
	marker.GetDrawingAttributeControl().SetDepthRange(0.1f, 0.2f);
	marker.GetVisibilityControl().SetMarkers(true);
	marker.GetMaterialMappingControl().SetMarkerColor(RGBAColor(1, 0, 0, 1));
	marker.GetSelectabilityControl().SetEverything(Selectability::Value::Off);

	/* This block of code does the following:
	 * 1. Sets up the selection options which will be used during the selection operation. In this case we decided
	 *		to select at the entity level, to sort the selection results from closer to most distant with respect 
	 *		to the camera, and to only consider the first selection element by setting the related limit to zero
	 * 2. Builds a key path. A key path is needed to correctly take into account cameras and matrix transforms 
	 *		when converting coordinates between systems. The key path is a collection of keys, going from the 
	 *		leaf to the root of the scene graph. The key path can omit some of the keys between the leaf and 
	 *		the root, but MUST contain all the include link between leaf and root 
	 * 3. Once the selection point has been properly converted to Window coordinate it is used together with
	 *		the selection options to carry out the selection operation by calling SelectByPoint */
	SelectionOptionsKit selectionOptions;
	SelectionResults selectionResults;
	Point selectionPoint;
	size_t numberOfSelectedItems;
	selectionOptions.SetLevel(Selection::Level::Entity).SetRelatedLimit(0).SetSorting(true);

	view.Update();

	KeyArray keyArray(9);
	keyArray[0] = marker;
	keyArray[1] = elements;
	keyArray[2] = modelKey;
	keyArray[3] = view.GetAttachedModelIncludeLink();
	keyArray[4] = view.GetSegmentKey();
	keyArray[5] = canvas.GetAttachedLayout().GetAttachedViewIncludeLink(0);
	keyArray[6] = canvas.GetAttachedLayout().GetSegmentKey();
	keyArray[7] = canvas.GetAttachedLayoutIncludeLink();
	keyArray[8] = wk;

	KeyPath keyPath;
	keyPath.SetKeys(keyArray);
	keyPath.ConvertCoordinate(Coordinate::Space::World, Point(-0.05f, -0.1f, 0), Coordinate::Space::Window, selectionPoint);

	numberOfSelectedItems = wk.GetSelectionControl().SelectByPoint(selectionPoint, selectionOptions, selectionResults);

	/* This block of code completes the following operations:
	 * 1. Obtains an iterator for the selection result list, and walks through it
	 * 2. For each selected item, it obtain its key, and then uses the key to obtain the
	 *		component selected, using the CADModel::GetComponentFromKey function
	 * 3. Obtains a Key Path for the selected component, and fully resolves it to the window
	 * 4. Highlights the selected component using its key path */
	SelectionResultsIterator it = selectionResults.GetIterator();
	HighlightOptionsKit hok;
	hok.SetStyleName("orangeHighlight");
	while(it.IsValid())
	{
		SelectionItem selected = it.GetItem();
		Key geometryKey;
		selected.ShowSelectedItem(geometryKey);
		Component selectedComponent = modelFile.GetComponentFromKey(geometryKey);
		KeyPathArray selectionKeyPath = Component::GetKeyPath(selectedComponent);

		wk.GetHighlightControl().Highlight(selectionKeyPath[0], hok);
		it.Next();
	}

	description.InsertText(Point(0, 0.90f, 0), "The component under the red cross");
	description.InsertText(Point(0, 0.85f, 0), "is selected and highlighted in orange");

	view.Update(Window::UpdateControl::CompleteUpdate);
	}
	catch(IOException e)
	{
		std::cout << e.what() << std::endl;
	}
}
